// Boost.Geometry

// This file was modified by Oracle on 2018.
// Modifications copyright (c) 2018, Oracle and/or its affiliates.
// Contributed and/or modified by Adam Wulkiewicz, on behalf of Oracle

// Use, modification and distribution is subject to the Boost Software License,
// Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_GEOMETRY_SRS_SHARED_GRIDS_HPP
#define BOOST_GEOMETRY_SRS_SHARED_GRIDS_HPP


#include <boost/geometry/srs/projections/impl/pj_apply_gridshift_shared.hpp>
#include <boost/geometry/srs/projections/impl/pj_gridlist_shared.hpp>
#include <boost/geometry/srs/projections/shared_grids.hpp>


#endif // BOOST_GEOMETRY_SRS_SHARED_GRIDS_HPP
