from .breaks import hinge15_breaks
from .breaks import hinge30_breaks
from .breaks import natural_breaks
from .breaks import quantile_breaks
from .breaks import percentile_breaks
from .breaks import stddev_breaks


